@@ -5,6 +5,8 @@
 
 #include <WServer>
 
+#include <qwbackend.h>
+
 #include <QObject>
 
 WAYLIB_SERVER_BEGIN_NAMESPACE
@@ -12,25 +14,36 @@ WAYLIB_SERVER_BEGIN_NAMESPACE
 class WOutput;
 class WInputDevice;
 class WBackendPrivate;
-class WBackend : public WServerInterface, public WObject
+class WAYLIB_SERVER_EXPORT WBackend : public QObject, public WObject,  public WServerInterface
 {
+    Q_OBJECT
     friend class WOutputPrivate;
     W_DECLARE_PRIVATE(WBackend)
+
 public:
-    WBackend();
+    explicit WBackend();
+
+    QW_NAMESPACE::qw_backend *handle() const;
 
     QVector<WOutput*> outputList() const;
     QVector<WInputDevice*> inputDeviceList() const;
 
-protected:
-    virtual void outputAdded(WOutput *output);
-    virtual void outputRemoved(WOutput *output);
-    virtual void inputAdded(WInputDevice *input);
-    virtual void inputRemoved(WInputDevice *input);
+    bool hasDrm() const;
+    bool hasX11() const;
+    bool hasWayland() const;
 
+Q_SIGNALS:
+    void outputAdded(WOutput *output);
+    void outputRemoved(WOutput *output);
+
+    void inputAdded(WInputDevice *input);
+    void inputRemoved(WInputDevice *input);
+
+protected:
     void create(WServer *server) override;
     void destroy(WServer *server) override;
     wl_global *global() const override;
+    QByteArrayView interfaceName() const override;
 };
 
 WAYLIB_SERVER_END_NAMESPACE
