@@ -6,28 +6,16 @@
 #include "wcursor.h"
 #include "wseat.h"
 #include "wtools.h"
-#include "wquickgammacontrol_p.h"
 #include "platformplugin/qwlrootscreen.h"
+#include "private/wglobal_p.h"
 
 #include <qwoutput.h>
 #include <qwoutputlayout.h>
 #include <qwrenderer.h>
 #include <qwswapchain.h>
 #include <qwallocator.h>
-
-extern "C" {
-#define static
-#include <wlr/types/wlr_output.h>
-#include <wlr/render/interface.h>
-#undef static
-#include <wlr/types/wlr_output_layout.h>
-#ifdef slots
-#undef slots // using at swapchain.h
-#endif
-#include <wlr/render/swapchain.h>
-#include <wlr/render/allocator.h>
-#include <wlr/render/wlr_renderer.h>
-}
+#include <qwrendererinterface.h>
+#include <qwoutputinterface.h>
 
 #include <QLoggingCategory>
 #include <QCoreApplication>
@@ -42,52 +30,26 @@ WAYLIB_SERVER_BEGIN_NAMESPACE
 
 Q_LOGGING_CATEGORY(qLcOutput, "waylib.server.output", QtWarningMsg)
 
-class Q_DECL_HIDDEN WOutputPrivate : public WObjectPrivate
+class Q_DECL_HIDDEN WOutputPrivate : public WWrapObjectPrivate
 {
 public:
-    WOutputPrivate(WOutput *qq, QWOutput *handle)
-        : WObjectPrivate(qq)
-        , handle(handle)
+    WOutputPrivate(WOutput *qq, qw_output *handle)
+        : WWrapObjectPrivate(qq)
     {
-        this->handle->setData(this, qq);
-
-        QObject::connect(this->handle.get(), qOverload<wlr_output_event_commit*>(&QWOutput::commit),
-                         qq, [qq] (wlr_output_event_commit *event) {
-            if (event->state->committed & WLR_OUTPUT_STATE_SCALE) {
-                Q_EMIT qq->scaleChanged();
-                Q_EMIT qq->effectiveSizeChanged();
-            }
-
-            if (event->state->committed & WLR_OUTPUT_STATE_MODE) {
-                Q_EMIT qq->modeChanged();
-                Q_EMIT qq->transformedSizeChanged();
-                Q_EMIT qq->effectiveSizeChanged();
-            }
-
-            if (event->state->committed & WLR_OUTPUT_STATE_TRANSFORM) {
-                Q_EMIT qq->orientationChanged();
-                Q_EMIT qq->transformedSizeChanged();
-                Q_EMIT qq->effectiveSizeChanged();
-            }
-
-            if (event->state->committed & WLR_OUTPUT_STATE_BUFFER)
-                Q_EMIT qq->bufferCommitted();
-        });
+        initHandle(handle);
+        this->handle()->set_data(this, qq);
     }
 
-    ~WOutputPrivate() {
-        handle->setData(this, nullptr);
+    void instantRelease() override {
+        handle()->set_data(nullptr, nullptr);
         if (layout)
             layout->remove(q_func());
     }
 
-    inline wlr_output *nativeHandle() const {
-        Q_ASSERT(handle);
-        return handle->handle();
-    };
+    WWRAP_HANDLE_FUNCTIONS(qw_output, wlr_output)
 
     inline QSize size() const {
-        Q_ASSERT(handle);
+        Q_ASSERT(handle());
         return QSize(nativeHandle()->width, nativeHandle()->height);
     }
 
@@ -97,7 +59,6 @@ public:
 
     W_DECLARE_PUBLIC(WOutput)
 
-    QPointer<QWOutput> handle;
     bool forceSoftwareCursor = false;
     QWlrootsScreen *screen = nullptr;
     QQuickWindow *window = nullptr;
@@ -106,11 +67,35 @@ public:
     WOutputLayout *layout = nullptr;
 };
 
-WOutput::WOutput(QWOutput *handle, WBackend *backend)
-    : QObject()
-    , WObject(*new WOutputPrivate(this, handle))
+WOutput::WOutput(qw_output *handle, WBackend *backend)
+    : WWrapObject(*new WOutputPrivate(this, handle))
 {
     d_func()->backend = backend;
+    connect(handle, qOverload<wlr_output_event_commit*>(&qw_output::notify_commit),
+            this, [this] (wlr_output_event_commit *event) {
+        if (event->state->committed & WLR_OUTPUT_STATE_SCALE) {
+            Q_EMIT this->scaleChanged();
+            Q_EMIT this->effectiveSizeChanged();
+        }
+
+        if (event->state->committed & WLR_OUTPUT_STATE_MODE) {
+            Q_EMIT this->modeChanged();
+            Q_EMIT this->transformedSizeChanged();
+            Q_EMIT this->effectiveSizeChanged();
+        }
+
+        if (event->state->committed & WLR_OUTPUT_STATE_TRANSFORM) {
+            Q_EMIT this->orientationChanged();
+            Q_EMIT this->transformedSizeChanged();
+            Q_EMIT this->effectiveSizeChanged();
+        }
+
+        if (event->state->committed & WLR_OUTPUT_STATE_BUFFER)
+            Q_EMIT this->bufferCommitted();
+
+        if (event->state->committed & WLR_OUTPUT_STATE_ENABLED)
+            Q_EMIT this->enabledChanged();
+    });
 }
 
 WOutput::~WOutput()
@@ -130,22 +115,22 @@ WServer *WOutput::server() const
     return d->backend->server();
 }
 
-QWRenderer *WOutput::renderer() const
+qw_renderer *WOutput::renderer() const
 {
     W_DC(WOutput);
-    return QWRenderer::from(d->nativeHandle()->renderer);
+    return qw_renderer::from(d->nativeHandle()->renderer);
 }
 
-QWSwapchain *WOutput::swapchain() const
+qw_swapchain *WOutput::swapchain() const
 {
     W_DC(WOutput);
-    return QWSwapchain::from(d->nativeHandle()->swapchain);
+    return qw_swapchain::from(d->nativeHandle()->swapchain);
 }
 
-QWAllocator *WOutput::allocator() const
+qw_allocator *WOutput::allocator() const
 {
     W_DC(WOutput);
-    return QWAllocator::from(d->nativeHandle()->allocator);
+    return qw_allocator::from(d->nativeHandle()->allocator);
 }
 
 // Copy from wlroots
@@ -390,34 +375,79 @@ static bool wlr_output_configure_primary_swapchain(struct wlr_output *output, in
     *swapchain_ptr = swapchain;
     return true;
 }
+
+static bool output_pick_cursor_format(struct wlr_output *output,
+                                      struct wlr_drm_format *format,
+                                      uint32_t drm_format) {
+    struct wlr_allocator *allocator = output->allocator;
+    assert(allocator != NULL);
+
+    const struct wlr_drm_format_set *display_formats = NULL;
+    if (output->impl->get_cursor_formats) {
+        display_formats =
+            output->impl->get_cursor_formats(output, allocator->buffer_caps);
+        if (display_formats == NULL) {
+            qCDebug(qLcOutput, "Failed to get cursor display formats");
+            return false;
+        }
+    }
+
+    return output_pick_format(output, display_formats, format, drm_format);
+}
 // End
 
-bool WOutput::configureSwapchain(const QSize &size, uint32_t format,
-                                 QWSwapchain **swapchain, bool doTest)
+bool WOutput::configurePrimarySwapchain(const QSize &size, uint32_t format,
+                                        qw_swapchain **swapchain, bool doTest)
 {
+    Q_ASSERT(!size.isEmpty());
     wlr_swapchain *sc = (*swapchain)->handle();
     bool ok = wlr_output_configure_primary_swapchain(nativeHandle(), size.width(), size.height(),
                                                      format, &sc, doTest);
     if (!ok)
         return false;
-    *swapchain = QWSwapchain::from(sc);
+    *swapchain = qw_swapchain::from(sc);
     return true;
 }
 
-QWOutput *WOutput::handle() const
+bool WOutput::configureCursorSwapchain(const QSize &size, uint32_t drmFormat, qw_swapchain **swapchain)
+{
+    Q_ASSERT(!size.isEmpty());
+    auto sc = *swapchain;
+    if (!sc || sc->handle()->width != size.width() || sc->handle()->height != size.height()) {
+        wlr_drm_format format = {0};
+        if (!output_pick_cursor_format(nativeHandle(), &format, drmFormat)) {
+            qCDebug(qLcOutput, "Failed to pick cursor format");
+            return false;
+        }
+
+        delete sc;
+        sc = qw_swapchain::create(*allocator(), size.width(), size.height(), &format);
+        wlr_drm_format_finish(&format);
+        if (!sc) {
+            qCDebug(qLcOutput, "Failed to create cursor swapchain");
+            return false;
+        }
+    }
+
+    *swapchain = sc;
+    return true;
+}
+
+qw_output *WOutput::handle() const
 {
     W_DC(WOutput);
-    return d->handle.data();
+    return d->handle();
 }
 
 wlr_output *WOutput::nativeHandle() const
 {
-    return handle()->handle();
+    W_DC(WOutput);
+    return d->nativeHandle();
 }
 
-WOutput *WOutput::fromHandle(const QWOutput *handle)
+WOutput *WOutput::fromHandle(const qw_output *handle)
 {
-    return handle->getData<WOutput>();
+    return handle->get_data<WOutput>();
 }
 
 WOutput *WOutput::fromScreen(const QScreen *screen)
@@ -437,6 +467,18 @@ QWlrootsScreen *WOutput::screen() const
     return d->screen;
 }
 
+QString WOutput::name() const
+{
+    W_DC(WOutput);
+    return QString::fromUtf8(d->nativeHandle()->name);
+}
+
+bool WOutput::isEnabled() const
+{
+    W_DC(WOutput);
+    return d->nativeHandle()->enabled;
+}
+
 QPoint WOutput::position() const
 {
     W_DC(WOutput);
@@ -446,7 +488,7 @@ QPoint WOutput::position() const
     if (Q_UNLIKELY(!d->layout))
         return p;
 
-    auto l_output = d->layout->get(d->handle);
+    auto l_output = d->layout->handle()->get(d->nativeHandle());
 
     if (Q_UNLIKELY(!l_output))
         return p;
@@ -464,15 +506,18 @@ QSize WOutput::size() const
 QSize WOutput::transformedSize() const
 {
     W_DC(WOutput);
-
-    return d->handle->transformedResolution();
+    int width, height;
+    d->handle()->transformed_resolution(&width, &height);
+    return QSize( width, height );
 }
 
 QSize WOutput::effectiveSize() const
 {
     W_DC(WOutput);
 
-    return d->handle->effectiveResolution();
+    int width, height;
+    d->handle()->effective_resolution(&width, &height);
+    return QSize( width, height );
 }
 
 WOutput::Transform WOutput::orientation() const
@@ -489,28 +534,6 @@ float WOutput::scale() const
     return d->nativeHandle()->scale;
 }
 
-QImage::Format WOutput::preferredReadFormat() const
-{
-    W_DC(WOutput);
-
-    auto renderer = d->nativeHandle()->renderer;
-    // ###: The wlr_output_preferred_read_format force request
-    // attach the renderer to wlr_output, but maybe the renderer
-    // is rendering, you will get a crash at renderer_bind_buffer.
-    // So, if it's rendering, we direct get the preferred read
-    // format of current buffer by renderer->impl->preferred_read_format.
-    if (renderer && renderer->rendering) {
-        if (!renderer->impl->preferred_read_format
-            || !renderer->impl->read_pixels) {
-            return QImage::Format_Invalid;
-        }
-
-        return WTools::toImageFormat(renderer->impl->preferred_read_format(renderer));
-    }
-
-    return WTools::toImageFormat(d->handle->preferredReadFormat());
-}
-
 void WOutput::attach(QQuickWindow *window)
 {
     W_D(WOutput);
@@ -543,86 +566,22 @@ WOutputLayout *WOutput::layout() const
 void WOutput::addCursor(WCursor *cursor)
 {
     static_cast<QWlrootsCursor*>(screen()->cursor())->addCursor(cursor);
+    Q_EMIT cursorAdded(cursor);
+    Q_EMIT cursorListChanged();
 }
 
 void WOutput::removeCursor(WCursor *cursor)
 {
     static_cast<QWlrootsCursor*>(screen()->cursor())->removeCursor(cursor);
+    Q_EMIT cursorRemoved(cursor);
+    Q_EMIT cursorListChanged();
 }
 
-QList<WCursor *> WOutput::cursorList() const
+const QList<WCursor *> &WOutput::cursorList() const
 {
     return static_cast<QWlrootsCursor*>(screen()->cursor())->cursors;
 }
 
-bool WOutput::setGammaLut(size_t ramp_size, uint16_t* r, uint16_t* g, uint16_t* b)
-{
-    W_D(WOutput);
-    wlr_output_state state;
-
-    wlr_output_state_init(&state);
-    if (!wlr_output_state_set_gamma_lut(&state, ramp_size, r, g, b)) {
-        wlr_output_state_finish(&state);
-        qCWarning(qLcOutput) << "Gamma lut can't set to state!";
-        return false;
-    }
-
-    if (!handle()->testState(&state)) {
-        wlr_output_state_finish(&state);
-        qCWarning(qLcOutput) << "The gamma lut state can't accepted by the backend!";
-        return false;
-    }
-
-    bool ok = handle()->commitState(&state);
-    wlr_output_state_finish(&state);
-    if (!ok)
-        qCWarning(qLcOutput) << "Output commitState failed!";
-    return ok;
-}
-
-bool WOutput::enable(bool enabled)
-{
-    W_D(WOutput);
-    d->handle->enable(enabled);
-    return true;
-}
-
-void WOutput::enableAdaptiveSync(bool enabled)
-{
-    W_D(WOutput);
-    d->handle->enableAdaptiveSync(enabled);
-}
-
-void WOutput::setMode(wlr_output_mode *mode)
-{
-    W_D(WOutput);
-    d->handle->setMode(mode);
-}
-
-void WOutput::setCustomMode(const QSize &size, int32_t refresh)
-{
-    W_D(WOutput);
-    d->handle->setCustomMode(size, refresh);
-}
-
-bool WOutput::test()
-{
-    W_D(WOutput);
-    return d->handle->test();
-}
-
-bool WOutput::commit()
-{
-    W_D(WOutput);
-    return d->handle->commit();
-}
-
-void WOutput::rollback()
-{
-    W_D(WOutput);
-    d->handle->rollback();
-}
-
 bool WOutput::forceSoftwareCursor() const
 {
     W_DC(WOutput);
@@ -635,7 +594,7 @@ void WOutput::setForceSoftwareCursor(bool on)
     if (d->forceSoftwareCursor == on)
         return;
     d->forceSoftwareCursor = on;
-    d->handle->lockSoftwareCursors(on);
+    d->handle()->lock_software_cursors(on);
 
     Q_EMIT forceSoftwareCursorChanged();
 }
