@@ -2,18 +2,21 @@
 // SPDX-License-Identifier: Apache-2.0 OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "wbufferrenderer_p.h"
-#include "wtexture.h"
 #include "wrenderhelper.h"
 #include "wqmlhelper_p.h"
 #include "wtools.h"
+#include "wsgtextureprovider.h"
 
 #include <qwbuffer.h>
 #include <qwtexture.h>
 #include <qwrenderer.h>
 #include <qwswapchain.h>
 #include <qwoutput.h>
+#include <qwallocator.h>
+#include <qwrendererinterface.h>
 
 #include <QSGImageNode>
+#include <QSGSimpleRectNode>
 
 #define protected public
 #define private public
@@ -31,19 +34,7 @@
 #include <private/qrhigles2_p.h>
 #include <private/qopenglcontext_p.h>
 #endif
-
-extern "C" {
-#include <wlr/types/wlr_damage_ring.h>
-#include <wlr/render/drm_format_set.h>
-#define static
-#include <wlr/render/wlr_renderer.h>
-#include <wlr/render/interface.h>
-#undef static
-#ifdef slots
-#undef slots // using at swapchain.h
-#endif
-#include <wlr/render/swapchain.h>
-}
+#include <private/qsgbatchrenderer_p.h>
 
 #include <pixman.h>
 #include <drm_fourcc.h>
@@ -52,7 +43,7 @@ extern "C" {
 QW_USE_NAMESPACE
 WAYLIB_SERVER_BEGIN_NAMESPACE
 
-struct PixmanRegion
+struct Q_DECL_HIDDEN PixmanRegion
 {
     PixmanRegion() {
         pixman_region32_init(&data);
@@ -79,7 +70,7 @@ inline static WImageRenderTarget *getImageFrom(const QQuickRenderTarget &rt)
     return static_cast<WImageRenderTarget*>(d->u.paintDevice);
 }
 
-static const wlr_drm_format *pickFormat(QWRenderer *renderer, uint32_t format)
+static const wlr_drm_format *pickFormat(qw_renderer *renderer, uint32_t format)
 {
     auto r = renderer->handle();
     if (!r->impl->get_render_formats) {
@@ -109,92 +100,17 @@ static void applyTransform(QSGSoftwareRenderer *renderer, const QTransform &t)
     }
 }
 
-class Q_DECL_HIDDEN TextureProvider : public QSGTextureProvider
-{
-public:
-    explicit TextureProvider(WBufferRenderer *item)
-        : item(item) {}
-    ~TextureProvider() {
-        if (m_texture)
-            cleanTexture();
-    }
-
-    QSGTexture *texture() const override {
-        return m_texture ? &m_texture->texture : nullptr;
-    }
-
-    void setBuffer(QWBuffer *buffer) {
-        Q_ASSERT(item);
-
-        if (m_texture)
-            cleanTexture();
-
-        Q_ASSERT(!m_texture);
-        if (buffer)
-            m_texture = new Texture(item->window(), item->output()->renderer(), buffer);
-
-        Q_EMIT textureChanged();
-    }
-
-    void cleanTexture() {
-        Q_ASSERT(item);
-
-        class TextureCleanupJob : public QRunnable
-        {
-        public:
-            TextureCleanupJob(Texture *texture)
-                : texture(texture) { }
-            void run() override {
-                delete texture;
-            }
-            Texture *texture;
-        };
-
-        // Delay clean the textures on the next render after.
-        item->window()->scheduleRenderJob(new TextureCleanupJob(m_texture),
-                                          QQuickWindow::AfterSynchronizingStage);
-        m_texture = nullptr;
-    }
-
-    void invalidate() {
-        cleanTexture();
-        item = nullptr;
-
-        Q_EMIT textureChanged();
-    }
-
-    WBufferRenderer *item;
-
-    struct Texture {
-        Texture(QQuickWindow *window, QWRenderer *renderer, QWBuffer *buffer)
-        {
-            qwtexture = QWTexture::fromBuffer(renderer, buffer);
-            WTexture::makeTexture(qwtexture, &texture, window);
-            texture.setOwnsTexture(false);
-        }
-
-        ~Texture() {
-            delete qwtexture;
-        }
-
-        QWTexture *qwtexture;
-        QSGPlainTexture texture;
-    };
-
-    Texture *m_texture = nullptr;
-};
-
 WBufferRenderer::WBufferRenderer(QQuickItem *parent)
     : QQuickItem(parent)
-    , m_cacheBuffer(false)
-    , m_forceCacheBuffer(false)
+    , m_cacheBuffer(true)
     , m_hideSource(false)
 {
-    m_textureProvider.reset(new TextureProvider(this));
+
 }
 
 WBufferRenderer::~WBufferRenderer()
 {
+    cleanTextureProvider();
     resetSources();
 
     delete m_renderHelper;
@@ -224,7 +140,7 @@ QList<QQuickItem*> WBufferRenderer::sourceList() const
     QList<QQuickItem*> list;
     list.reserve(m_sourceList.size());
 
-    for (const Data &i : m_sourceList)
+    for (const Data &i : std::as_const(m_sourceList))
         list.append(i.source);
 
     return list;
@@ -249,7 +165,7 @@ void WBufferRenderer::setSourceList(QList<QQuickItem*> sources, bool hideSource)
     m_sourceList.clear();
     m_hideSource = hideSource;
 
-    for (auto s : sources) {
+    for (auto s : std::as_const(sources)) {
         m_sourceList.append({s, nullptr});
 
         if (isRootItem(s))
@@ -284,22 +200,63 @@ void WBufferRenderer::setCacheBuffer(bool newCacheBuffer)
     Q_EMIT cacheBufferChanged();
 }
 
+void WBufferRenderer::lockCacheBuffer(QObject *owner)
+{
+    if (m_cacheBufferLocker.contains(owner))
+        return;
+    m_cacheBufferLocker.append(owner);
+    connect(owner, &QObject::destroyed, this, [this] {
+        unlockCacheBuffer(sender());
+    });
+    updateTextureProvider();
+}
+
+void WBufferRenderer::unlockCacheBuffer(QObject *owner)
+{
+    auto ok = m_cacheBufferLocker.removeOne(owner);
+    Q_ASSERT(ok);
+    ok = disconnect(owner, &QObject::destroyed, this, nullptr);
+    Q_ASSERT(ok);
+    updateTextureProvider();
+}
+
+QColor WBufferRenderer::clearColor() const
+{
+    return m_clearColor;
+}
+
+void WBufferRenderer::setClearColor(const QColor &clearColor)
+{
+    m_clearColor = clearColor;
+}
+
 QSGRenderer *WBufferRenderer::currentRenderer() const
 {
     return state.renderer;
 }
 
+QSGBatchRenderer::Renderer *WBufferRenderer::currentBatchRenderer() const
+{
+    Q_ASSERT(state.renderer == state.batchRenderer);
+    return state.batchRenderer;
+}
+
+qreal WBufferRenderer::currentDevicePixelRatio() const
+{
+    return state.devicePixelRatio;
+}
+
 const QMatrix4x4 &WBufferRenderer::currentWorldTransform() const
 {
     return state.worldTransform;
 }
 
-QWBuffer *WBufferRenderer::currentBuffer() const
+qw_buffer *WBufferRenderer::currentBuffer() const
 {
     return state.buffer;
 }
 
-QWBuffer *WBufferRenderer::lastBuffer() const
+qw_buffer *WBufferRenderer::lastBuffer() const
 {
     return m_lastBuffer;
 }
@@ -310,12 +267,12 @@ QRhiTexture *WBufferRenderer::currentRenderTarget() const
     return textureRT->description().colorAttachmentAt(0)->texture();
 }
 
-const QWDamageRing *WBufferRenderer::damageRing() const
+const qw_damage_ring *WBufferRenderer::damageRing() const
 {
     return &m_damageRing;
 }
 
-QWDamageRing *WBufferRenderer::damageRing()
+qw_damage_ring *WBufferRenderer::damageRing()
 {
     return &m_damageRing;
 }
@@ -327,16 +284,65 @@ bool WBufferRenderer::isTextureProvider() const
 
 QSGTextureProvider *WBufferRenderer::textureProvider() const
 {
+    return wTextureProvider();
+}
+
+WSGTextureProvider *WBufferRenderer::wTextureProvider() const
+{
+    auto w = qobject_cast<WOutputRenderWindow*>(window());
+    auto d = QQuickItemPrivate::get(this);
+    if (!w || !d->sceneGraphRenderContext() || QThread::currentThread() != d->sceneGraphRenderContext()->thread()) {
+        qWarning("WBufferRenderer::textureProvider: can only be queried on the rendering thread of an WOutputRenderWindow");
+        return nullptr;
+    }
+
+    if (!m_textureProvider) {
+        m_textureProvider.reset(new WSGTextureProvider(w));
+        m_textureProvider->setBuffer(m_lastBuffer);
+    }
+
     return m_textureProvider.get();
 }
 
-QWBuffer *WBufferRenderer::beginRender(const QSize &pixelSize, qreal devicePixelRatio,
-                                       uint32_t format, RenderFlags flags)
+QTransform WBufferRenderer::inputMapToOutput(const QRectF &sourceRect, const QRectF &targetRect,
+                                             const QSize &pixelSize, const qreal devicePixelRatio)
+{
+    Q_ASSERT(pixelSize.isValid());
+
+    QTransform t;
+    const auto outputSize = QSizeF(pixelSize) / devicePixelRatio;
+
+    if (sourceRect.isValid())
+        t.translate(-sourceRect.x(), -sourceRect.y());
+    if (targetRect.isValid())
+        t.translate(targetRect.x(), targetRect.y());
+
+    if (sourceRect.isValid()) {
+        t.scale(outputSize.width() / sourceRect.width(),
+                outputSize.height() / sourceRect.height());
+    }
+
+    if (targetRect.isValid()) {
+        t.scale(targetRect.width() / outputSize.width(),
+                targetRect.height() / outputSize.height());
+    }
+
+    return t;
+}
+
+qw_buffer *WBufferRenderer::beginRender(const QSize &pixelSize, qreal devicePixelRatio,
+                                        uint32_t format, RenderFlags flags)
 {
     Q_ASSERT(!state.buffer);
     Q_ASSERT(m_output);
 
-    m_damageRing.setBounds(pixelSize);
+    if (pixelSize.isEmpty())
+        return nullptr;
+
+    Q_EMIT beforeRendering();
+
+    m_damageRing.set_bounds(pixelSize.width(), pixelSize.height());
+
     // configure swapchain
     if (flags.testFlag(RenderFlag::DontConfigureSwapchain)) {
         auto renderFormat = pickFormat(m_output->renderer(), format);
@@ -344,26 +350,30 @@ QWBuffer *WBufferRenderer::beginRender(const QSize &pixelSize, qreal devicePixel
             qWarning("wlr_renderer doesn't support format 0x%s", drmGetFormatName(format));
             return nullptr;
         }
+
         if (!m_swapchain || QSize(m_swapchain->handle()->width, m_swapchain->handle()->height) != pixelSize
             || m_swapchain->handle()->format.format != renderFormat->format) {
             if (m_swapchain)
                 delete m_swapchain;
-            m_swapchain = QWSwapchain::create(m_output->allocator(), pixelSize, renderFormat);
+            m_swapchain = qw_swapchain::create(m_output->allocator()->handle(), pixelSize.width(), pixelSize.height(), renderFormat);
         }
+    } else if (flags.testFlag(RenderFlag::UseCursorFormats)) {
+        bool ok = m_output->configureCursorSwapchain(pixelSize, format, &m_swapchain);
+        if (!ok)
+            return nullptr;
     } else {
-        bool ok = m_output->configureSwapchain(pixelSize, format, &m_swapchain, !flags.testFlag(DontTestSwapchain));
+        bool ok = m_output->configurePrimarySwapchain(pixelSize, format, &m_swapchain,
+                                                      !flags.testFlag(DontTestSwapchain));
         if (!ok)
             return nullptr;
     }
 
-    if (!shouldCacheBuffer())
-        m_textureProvider->setBuffer(nullptr);
-
     // TODO: Support scanout buffer of wlr_surface(from WSurfaceItem)
     int bufferAge;
-    auto buffer = m_swapchain->acquire(&bufferAge);
-    if (!buffer)
+    auto wbuffer = m_swapchain->acquire(&bufferAge);
+    if (!wbuffer)
         return nullptr;
+    auto buffer = qw_buffer::from(wbuffer);
 
     if (!m_renderHelper)
         m_renderHelper = new WRenderHelper(m_output->renderer());
@@ -383,7 +393,19 @@ QWBuffer *WBufferRenderer::beginRender(const QSize &pixelSize, qreal devicePixel
 
     if (rtd->type == QQuickRenderTargetPrivate::Type::PaintDevice) {
         sgRT.paintDevice = rtd->u.paintDevice;
+
+        // // For software renderer, update the dirty parts relative to the last paint device.
+        PixmanRegion damage;
+        m_damageRing.get_buffer_damage(bufferAge, damage);
+        state.dirty = WTools::fromPixmanRegion(damage);
+
+        if (devicePixelRatio != 1.0) {
+            state.dirty = QTransform::fromScale(1.0 / devicePixelRatio,
+                                                1.0 / devicePixelRatio).map(state.dirty);
+        }
     } else {
+        state.dirty = QRegion();
+
         Q_ASSERT(rtd->type == QQuickRenderTargetPrivate::Type::RhiRenderTarget);
         sgRT.rt = rtd->u.rhiRt;
         sgRT.cb = wd->redirect.commandBuffer;
@@ -393,7 +415,7 @@ QWBuffer *WBufferRenderer::beginRender(const QSize &pixelSize, qreal devicePixel
 #ifndef QT_NO_OPENGL
         if (wd->rhi->backend() == QRhi::OpenGLES2) {
             auto glRT = QRHI_RES(QGles2TextureRenderTarget, rtd->u.rhiRt);
-            Q_ASSERT(glRT->framebuffer > 0);
+            Q_ASSERT(glRT->framebuffer >= 0);
             auto glContext = QOpenGLContext::currentContext();
             Q_ASSERT(glContext);
             QOpenGLContextPrivate::get(glContext)->defaultFboRedirect = glRT->framebuffer;
@@ -406,37 +428,84 @@ QWBuffer *WBufferRenderer::beginRender(const QSize &pixelSize, qreal devicePixel
     state.pixelSize = pixelSize;
     state.devicePixelRatio = devicePixelRatio;
     state.bufferAge = bufferAge;
-    state.lastRT = lastRT;
     state.buffer = buffer;
     state.renderTarget = rt;
     state.sgRenderTarget = sgRT;
 
-    if (!shouldCacheBuffer())
-        m_textureProvider->setBuffer(buffer);
-
     return buffer;
 }
 
-void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix, bool preserveColorContents)
+inline static QRect scaleToRect(const QRectF &s, qreal scale) {
+    return QRect((s.topLeft() * scale).toPoint(),
+                 (s.size() * scale).toSize());
+}
+
+void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix,
+                             const QRectF &sourceRect, const QRectF &targetRect,
+                             bool preserveColorContents)
 {
     Q_ASSERT(state.buffer);
 
     const auto &source = m_sourceList.at(sourceIndex);
-    QSGRenderer *renderer = ensureRenderer(source.source, state.context);
+    QSGRenderer *renderer = ensureRenderer(sourceIndex, state.context);
     auto wd = QQuickWindowPrivate::get(window());
 
     const qreal devicePixelRatio = state.devicePixelRatio;
     state.renderer = renderer;
+    state.batchRenderer = dynamic_cast<QSGBatchRenderer::Renderer*>(renderer);
     state.worldTransform = renderMatrix;
-    state.worldTransform.optimize();
-    renderer->setDevicePixelRatio(devicePixelRatio);
+    // The renderer should always receive the window's DPR (Device Pixel Ratio)
+    // because, regardless of the DPR used for rendering, all resources within
+    // a window are loaded based on the window's own DPR.
+
+    // During rendering, certain specialized nodes (e.g., QSGCurveStrokeMaterialShader)
+    // use QSGRenderer::devicePixelRatio for specific calculations related to
+    // ShapePath::fillItem. When displaying this fillItem using Shape, it might
+    // be desirable for the QSGTexture provided by fillItem to scale completely
+    // to match the size of the Shape, regardless of its original size.
+
+    // If a PathRectangle is used, its width is set to
+    // textureSize.width / QQuickWindow::effectiveDevicePixelRatio. Here, the
+    // devicePixelRatio value is utilized because the width and height passed
+    // to PathRectangle need to be converted from pixel values to pixel-independent sizes.
+
+    // Returning to the issue mentioned earlier, QSGCurveStrokeMaterialShader
+    // uses QSGRenderer::devicePixelRatio for additional calculations that influence
+    // how Shape fills the QSGTexture provided by fillItem. Therefore, we need to ensure
+    // that QSGRenderer::devicePixelRatio and QQuickWindow::effectiveDevicePixelRatio
+    // are always consistent. Otherwise, some Items might render incorrectly.
+    renderer->setDevicePixelRatio(window()->effectiveDevicePixelRatio());
     renderer->setDeviceRect(QRect(QPoint(0, 0), state.pixelSize));
-    renderer->setViewportRect(state.pixelSize);
     renderer->setRenderTarget(state.sgRenderTarget);
+    const auto viewportRect = scaleToRect(targetRect, devicePixelRatio);
 
     auto softwareRenderer = dynamic_cast<QSGSoftwareRenderer*>(renderer);
     { // before render
         if (softwareRenderer) {
+            // Avoid do clear before paint, for the software renderer this
+            // work is expensive.
+            if (m_clearColor.alpha() == 0)
+                preserveColorContents = true;
+#if QT_VERSION >= QT_VERSION_CHECK(6, 9, 0)
+            softwareRenderer->setClearColorEnabled(!preserveColorContents);
+#else
+            auto bn = softwareRenderer->renderableNode(softwareRenderer->m_background);
+            if (bn) {
+                bn->m_opacity = preserveColorContents ? 0 : 1;
+            }
+#endif
+            if (!state.dirty.isEmpty()) {
+                softwareRenderer->m_dirtyRegion += state.dirty;
+                state.dirty = QRegion();
+            }
+
+            // because software renderer don't supports viewportRect,
+            // so use transform to simulation.
+            const auto mapTransform = inputMapToOutput(sourceRect, targetRect,
+                                                       state.pixelSize, state.devicePixelRatio);
+            if (!mapTransform.isIdentity())
+                state.worldTransform = mapTransform * state.worldTransform;
+            state.worldTransform.optimize();
             auto image = getImageFrom(state.renderTarget);
             image->setDevicePixelRatio(devicePixelRatio);
 
@@ -456,11 +525,24 @@ void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix, bo
                 applyTransform(softwareRenderer, t);
             }
         } else {
+            state.worldTransform.optimize();
+
             bool flipY = wd->rhi ? !wd->rhi->isYUpInNDC() : false;
             if (state.renderTarget.mirrorVertically())
                 flipY = !flipY;
 
-            QRectF rect(QPointF(0, 0), QSizeF(state.pixelSize) / devicePixelRatio);
+            if (viewportRect.isValid()) {
+                QRect vr = viewportRect;
+                if (flipY)
+                    vr.moveTop(-vr.y() + state.pixelSize.height() - vr.height());
+                renderer->setViewportRect(vr);
+            } else {
+                renderer->setViewportRect(QRect(QPoint(0, 0), state.pixelSize));
+            }
+
+            QRectF rect = sourceRect;
+            if (!rect.isValid())
+                rect = QRectF(QPointF(0, 0), QSizeF(state.pixelSize) / devicePixelRatio);
 
             const float left = rect.x();
             const float right = rect.x() + rect.width();
@@ -501,7 +583,7 @@ void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix, bo
     { // after render
         if (!softwareRenderer) {
             // TODO: get damage area from QRhi renderer
-            m_damageRing.addWhole();
+            m_damageRing.add_whole();
             // ###: maybe Qt bug? Before executing QRhi::endOffscreenFrame, we may
             // use the same QSGRenderer for multiple drawings. This can lead to
             // rendering the same content for different QSGRhiRenderTarget instances
@@ -512,6 +594,8 @@ void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix, bo
             // drawing result is available for use.
             wd->rhi->finish();
         } else {
+            state.dirty = softwareRenderer->flushRegion();
+
             auto currentImage = getImageFrom(state.renderTarget);
             Q_ASSERT(currentImage && currentImage == softwareRenderer->m_rt.paintDevice);
             currentImage->setDevicePixelRatio(1.0);
@@ -523,27 +607,19 @@ void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix, bo
 
             {
                 PixmanRegion damage;
-                m_damageRing.getBufferDamage(state.bufferAge, damage);
+                m_damageRing.get_buffer_damage(state.bufferAge, damage);
 
-                if (!damage.isEmpty() && state.lastRT.first != state.buffer && !state.lastRT.second.isNull()) {
-                    auto image = getImageFrom(state.lastRT.second);
-                    Q_ASSERT(image);
-                    Q_ASSERT(image->size() == state.pixelSize);
+                if (viewportRect.isValid()) {
+                    QRect imageRect = (currentImage->operator const QImage &()).rect();
+                    QRegion invalidRegion(imageRect);
+                    invalidRegion -= viewportRect;
+                    if (!scaledFlushRegion.isEmpty())
+                        invalidRegion &= scaledFlushRegion;
 
-                    // TODO: Don't use the previous render target, we can get the damage region of QtQuick
-                    // before QQuickRenderControl::render for QWDamageRing, and add dirty region to
-                    // QSGAbstractSoftwareRenderer to force repaint the damage region of current render target.
-                    QPainter pa(currentImage);
-
-                    PixmanRegion remainderDamage;
-                    ok = pixman_region32_subtract(remainderDamage, damage, scaledFlushDamage);
-                    Q_ASSERT(ok);
-
-                    int count = 0;
-                    auto rects = pixman_region32_rectangles(remainderDamage, &count);
-                    for (int i = 0; i < count; ++i) {
-                        auto r = rects[i];
-                        pa.drawImage(r.x1, r.y1, *image, r.x1, r.y1, r.x2 - r.x1, r.y2 - r.y1);
+                    if (!invalidRegion.isEmpty()) {
+                        QPainter pa(currentImage);
+                        for (const auto r : std::as_const(invalidRegion))
+                            pa.fillRect(r, softwareRenderer->clearColor());
                     }
                 }
             }
@@ -558,6 +634,9 @@ void WBufferRenderer::render(int sourceIndex, const QMatrix4x4 &renderMatrix, bo
         QRhiResourceUpdateBatch *resourceUpdates = wd->rhi->nextResourceUpdateBatch();
         dr->currentFrameCommandBuffer()->resourceUpdate(resourceUpdates);
     }
+
+    if (shouldCacheBuffer())
+        wTextureProvider()->setBuffer(state.buffer);
 }
 
 void WBufferRenderer::endRender()
@@ -566,12 +645,12 @@ void WBufferRenderer::endRender()
     auto buffer = state.buffer;
     state.buffer = nullptr;
     state.renderer = nullptr;
-    if (shouldCacheBuffer())
-        m_textureProvider->setBuffer(buffer);
+    state.batchRenderer = nullptr;
+    state.dirty = QRegion();
 
     m_lastBuffer = buffer;
     m_damageRing.rotate();
-    m_swapchain->setBufferSubmitted(buffer);
+    m_swapchain->set_buffer_submitted(*buffer);
     buffer->unlock();
 
 #ifndef QT_NO_OPENGL
@@ -583,6 +662,8 @@ void WBufferRenderer::endRender()
         QOpenGLContextPrivate::get(glContext)->defaultFboRedirect = GL_NONE;
     }
 #endif
+
+    Q_EMIT afterRendering();
 }
 
 void WBufferRenderer::componentComplete()
@@ -590,14 +671,6 @@ void WBufferRenderer::componentComplete()
     QQuickItem::componentComplete();
 }
 
-void WBufferRenderer::setForceCacheBuffer(bool force)
-{
-    if (m_forceCacheBuffer == force)
-        return;
-    m_forceCacheBuffer = force;
-    updateTextureProvider();
-}
-
 void WBufferRenderer::resetTextureProvider()
 {
     if (m_textureProvider)
@@ -606,7 +679,10 @@ void WBufferRenderer::resetTextureProvider()
 
 void WBufferRenderer::updateTextureProvider()
 {
-    if (shouldCacheBuffer()) {
+    if (!m_textureProvider)
+        return;
+
+    if (shouldCacheBuffer() && m_textureProvider->qwBuffer() != m_lastBuffer) {
         m_textureProvider->setBuffer(m_lastBuffer);
     } else {
         m_textureProvider->setBuffer(nullptr);
@@ -641,6 +717,11 @@ void WBufferRenderer::invalidateSceneGraph()
 }
 
 void WBufferRenderer::releaseResources()
+{
+    cleanTextureProvider();
+}
+
+void WBufferRenderer::cleanTextureProvider()
 {
     if (m_textureProvider) {
         class TextureProviderCleanupJob : public QRunnable
@@ -673,7 +754,9 @@ void WBufferRenderer::removeSource(int index)
     if (isRootItem(s.source))
         return;
 
-    s.renderer->deleteLater();
+    // Renderer of source is delay initialized in ensureRenderer. It might be null here.
+    if (s.renderer)
+        s.renderer->deleteLater();
     auto d = QQuickItemPrivate::get(s.source);
     if (d->inDestructor)
         return;
@@ -693,19 +776,16 @@ int WBufferRenderer::indexOfSource(QQuickItem *s)
     return -1;
 }
 
-QSGRenderer *WBufferRenderer::ensureRenderer(QQuickItem *source, QSGRenderContext *rc)
+QSGRenderer *WBufferRenderer::ensureRenderer(int sourceIndex, QSGRenderContext *rc)
 {
-    if (isRootItem(source))
+    Data &d = m_sourceList[sourceIndex];
+    if (isRootItem(d.source))
         return QQuickWindowPrivate::get(window())->renderer;
 
-    const int index = indexOfSource(source);
-    Q_ASSERT(index >= 0);
-    Data &d = m_sourceList[index];
-
     if (Q_LIKELY(d.renderer))
         return d.renderer;
 
-    auto rootNode = WQmlHelper::getRootNode(source);
+    auto rootNode = WQmlHelper::getRootNode(d.source);
     Q_ASSERT(rootNode);
 
     auto dr = qobject_cast<QSGDefaultRenderContext*>(rc);
@@ -717,7 +797,7 @@ QSGRenderer *WBufferRenderer::ensureRenderer(QQuickItem *source, QSGRenderContex
     QObject::connect(d.renderer, &QSGRenderer::sceneGraphChanged,
                      this, &WBufferRenderer::sceneGraphChanged);
 
-    d.renderer->setClearColor(Qt::transparent);
+    d.renderer->setClearColor(m_clearColor);
 
     return d.renderer;
 }
