@@ -10,16 +10,23 @@
 #include <QRect>
 
 struct pixman_region32;
+typedef int wl_shm_format_t;
+
+QT_BEGIN_NAMESPACE
+class QQuickItem;
+QT_END_NAMESPACE
 
 WAYLIB_SERVER_BEGIN_NAMESPACE
 
-class WTools
+class WAYLIB_SERVER_EXPORT WTools
 {
 public:
     static QImage fromPixmanImage(void *image, void *data = nullptr);
     static QImage::Format toImageFormat(uint32_t drmFormat);
     static uint32_t toDrmFormat(QImage::Format format);
     static QImage::Format convertToDrmSupportedFormat(QImage::Format format);
+    static uint32_t shmToDrmFormat(wl_shm_format_t shmFmt);
+    static wl_shm_format_t drmToShmFormat(uint32_t drmFmt);
     static QRegion fromPixmanRegion(pixman_region32 *region);
     static bool toPixmanRegion(const QRegion &region, pixman_region32 *pixmanRegion);
     static QRect fromWLRBox(void *box);
