@@ -6,7 +6,7 @@
 
 #include <QQuickWindow>
 
-class CustomWindow : public QQuickWindow
+class Q_DECL_HIDDEN CustomWindow : public QQuickWindow
 {
     Q_OBJECT
     Q_PROPERTY(QWindow* parent READ parent WRITE setParent NOTIFY parentChanged FINAL)
@@ -18,6 +18,11 @@ public:
 
     QWindow *parent() const;
     void setParent(QWindow *newParent);
+    Q_INVOKABLE inline bool destroy() {
+        // can't close with parent, setParent(nullptr) destroys subsurf and make new xdgsurface
+        QQuickWindow::destroy();
+        return true;
+    }
 
 signals:
     void parentChanged();
